"""
Copyright: Digital Observatory 2023 <digitalobservatory@qut.edu.au>
Author: Mat Bettinson <mat.bettinson@qut.edu.au>, Boyd Nguyen <thaihoang.nguyen@qut.edu.au>
"""

import logging
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from faster_whisper import WhisperModel
from faster_whisper.transcribe import Segment

from .transcript import WhisperTranscript

logger = logging.getLogger(__name__)

MODEL_SIZE = [
    "tiny",
    "tiny.en",
    "base",
    "base.en",
    "small",
    "small.en",
    "medium",
    "medium.en",
    "large-v1",
    "large-v2",
    "large-v3",
    "large",
]

DEVICE = ["cpu", "cuda", "auto"]

COMPUTE_STYLE = [
    "int8",
    "int8_float32",
    "int8_float16",
    "int8_bfloat16",
    "int16",
    "float16",
    "bfloat16",
    "float32",
]


class TranscriptionError(Exception):
    pass


class BaseTranscriber(ABC):
    @abstractmethod
    def transcribe(self, file, **kwargs) -> Any:
        pass


class WhisperTranscriber(BaseTranscriber):
    def __init__(
        self,
        model_size: str = "small",
        cpu_threads: int = 4,
        device: str = "cpu",
        device_index: int | list[int] = 0,
        compute_style: str = "int8",
        num_workers: int = 1,
    ):
        """Initialise the Whisper model

        Parameters
        ----------
        model_size : str, optional
            Size of the model to use (tiny, tiny.en, base, base.en,
            small, small.en, medium, medium.en, large-v1, large-v2,
            large-v3, or large), by default "small"
        cpu_threads : int, optional
            Number of threads to use when running on CPU, by default 4
        device : str, optional
            Device to use for computation ("cpu", "cuda", "auto"), by default "cpu"
        device_index : int | list[int], optional
            Device ID to use.
            The model can also be loaded on multiple GPUs by passing a list of IDs
            (e.g. [0, 1, 2, 3]).
            In that case, multiple transcriptions can run in parallel
            when transcribe() is called from multiple Python threads
            (see also num_workers), by default 0
        compute_style : str, optional
            Type to use for computation ("int8", "int8_float32", "int8_float16",
            "int8_bfloat16", "int16", "float16",
            "bfloat16", "float32"), by default "int8"
        num_workers : int, optional
            When transcribe() is called from multiple Python threads,
            having multiple workers enables true parallelism when
            running the model, by default 1

        Raises
        ------
        ValueError
            If model_size is not one of the acceptable values
        ValueError
            If device is not one of the acceptable values
        ValueError
            If compute_style is not one of the acceptable values
        """
        if model_size not in MODEL_SIZE:
            raise ValueError(f"model_name must be one of {MODEL_SIZE}")
        if device not in DEVICE:
            raise ValueError(f"device must be one of {DEVICE}")
        if compute_style not in COMPUTE_STYLE:
            raise ValueError(f"compute_style must be one of {COMPUTE_STYLE}")

        self._model = WhisperModel(
            model_size_or_path=model_size,
            device=device,
            device_index=device_index,
            cpu_threads=cpu_threads,
            compute_type=compute_style,
            num_workers=num_workers,
        )

    def transcribe(
        self, file: str | Path, prompt: str | None = None, vad_filter=True
    ) -> WhisperTranscript:
        """Transcribe an audio

        Parameters
        ----------
        file : str | Path
            Path to the audio file
        prompt : str | None, optional
            Optional text string or iterable of token ids to provide as a
            prompt for the first window, by default None
        vad_filter : bool, optional
            Enable the voice activity detection (VAD) to filter out parts of the audio
            without speech., by default True

        Raises
        ------
        TranscriptionError
            If nothing is returned, raise TranscriptionError.
        """
        segments, info = self._model.transcribe(
            file, initial_prompt=prompt, vad_filter=vad_filter  # type: ignore
        )

        if not segments and not info:
            raise TranscriptionError("Nothing was transcribed.")

        logger.info(
            f"Transcription completed. Language: {info.language}. Probability: {info.language_probability}. Duration: {info.duration}"
        )

        return WhisperTranscript(segments)



"""
Copyright: Digital Observatory 2023 <digitalobservatory@qut.edu.au>
Author: Mat Bettinson <mat.bettinson@qut.edu.au>
"""

import logging
import random
import string
from datetime import timedelta
from os import path, remove
from pathlib import Path
from time import time
from typing import Optional

from yt_dlp import YoutubeDL

from .scraper import VideoDetails, get_video_details
from .transcriber import BaseTranscriber, WhisperTranscriber
from .transcript import BaseTranscript

logger = logging.getLogger()


def transcribe(
    url: str,
    transcriber: BaseTranscriber = WhisperTranscriber(),
    prompt: str | bool | None = None,
    **kwargs,
) -> BaseTranscript:
    """Transcribe a YouTube video

    The function first downloads the audio of the video,
    then pass to the transcriber to do the transcription.

    Parameters
    ----------
    url : str
        Full URL of the YouTube video
    transcriber : BaseTranscriber, optional
        Any subclass of BaseTranscriber, by default WhisperTranscriber()
    prompt : str | bool | None, optional
        The prompt to be put to the transcription model,
        if None, use video title and description as prompt,
        if False, prompt will not be added,
        by default None

    Returns
    -------
    BaseTranscript
    """
    meta = _get_metadata(url)

    if prompt is None:
        logger.info("No prompt provided, using video title and description as prompt")
        if meta:
            prompt = meta.title + ": " + meta.shortDescription
    elif prompt == False:
        logger.info("prompt is set to False. No prompt is used")
        prompt = None

    if meta and meta.title:
        file_name = f"{meta.videoId}.webm"
    else:
        random_name = "".join(random.choices(string.ascii_letters + string.digits, k=6))
        file_name = f"{random_name}.webm"

    _download_audio(url, file_name)

    start_time = time()
    transcript = transcriber.transcribe(file_name, prompt=prompt, **kwargs)

    logger.info(f"Transcription finished in {_get_duration(start_time, time())}")

    remove(file_name)

    return transcript


def _get_metadata(url: str) -> VideoDetails | None:
    start_time = time()
    video_details = get_video_details(url)

    if video_details is None:
        logger.warning("Failed to get video details")
        return None

    logger.info("Got metadata in" + str(timedelta(seconds=round(time() - start_time))))
    return video_details


def _download_audio(url: str, file: str | Path, opts: Optional[dict] = None):
    if not opts:
        ydl_opts = {"format": "worstaudio[ext=webm]", "outtmpl": "%(id)s.%(ext)s"}

    if not path.isfile(file):
        start_time = time()

        try:
            with YoutubeDL(ydl_opts) as ydl:
                ydl.download(url)
        except Exception as e:
            raise e

        logger.info(f"{file} downloaded in {_get_duration(start_time, time())}")
    else:
        raise FileExistsError(f"{Path(file).resolve()} already exists.")


def _get_duration(start, end):
    return str(timedelta(seconds=round(end - start)))



TINY = "tiny"
TINY_EN = "tiny.en"
BASE = "base"
BASE_EN = "base.en"
SMALL = "small"
SMALL_EN = "small.en"
MEDIUM = "medium"
MEDIUM_EN = "medium.en"
LARGE_V1 = "large-v1"
LARGE_V2 = "large-v2"
LARGE_V3 = "large-v3"
LARGE = "large"



"""
Copyright: Digital Observatory 2023 <digitalobservatory@qut.edu.au>
Author: Mat Bettinson <mat.bettinson@qut.edu.au>
"""

import json
import logging
import re
from dataclasses import dataclass

import requests
from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)


@dataclass
class VideoDetails:
    videoId: str
    title: str
    shortDescription: str
    lengthSeconds: int
    keywords: list[str] | None
    channelId: str
    author: str
    viewcount: int


def get_video_details(url: str) -> VideoDetails | None:
    # Send a HTTP request to the given URL
    res = requests.get(url)
    # Throw an error if status is not 200
    res.raise_for_status()

    # Parse HTML response
    soup = BeautifulSoup(res.text, "html.parser")

    # Find body tag and then find next sibling script tag
    body = soup.body
    assert body is not None
    script = body.find("script")

    # If script tag is not found, return None
    if script is None:
        logger.warning("Script tag not found")
        return None

    # Extract JavaScript object using regex
    match = re.search(r"\{.*\}", script.text)

    # If object is not found, return None
    if match is None:
        logger.warning("JavaScript object not found")
        return None

    # Decode the JavaScript object into a Python dictionary
    obj = json.loads(match.group(0))
    vd = obj["videoDetails"]
    return VideoDetails(
        videoId=vd["videoId"],
        title=vd["title"],
        shortDescription=vd["shortDescription"],
        lengthSeconds=int(vd["lengthSeconds"]),
        keywords=vd["keywords"] if "keywords" in vd else None,
        channelId=vd["channelId"],
        author=vd["author"],
        viewcount=int(vd["viewCount"]),
    )



import json
from abc import ABC, abstractmethod
from datetime import timedelta
from pathlib import Path
from typing import Iterable

from faster_whisper.transcribe import Segment


class BaseTranscript(ABC):
    @abstractmethod
    def text(self):
        pass

    @abstractmethod
    def json(self):
        pass

    @abstractmethod
    def srt(self):
        pass


class WhisperTranscript(BaseTranscript):
    def __init__(self, segments: Iterable[Segment]):
        self.segments = list(segments)

    def text(self) -> str:
        return " ".join([segment.text.strip() for segment in self.segments])

    def json(self) -> str:
        obj = []
        for segment in self.segments:
            obj.append(
                {
                    "id": segment.id,
                    "start": round(segment.start, 2),
                    "end": round(segment.end, 2),
                    "text": segment.text.strip(),
                }
            )
        return json.dumps(obj)

    def srt(self, file_path: str | Path | None = None) -> str:
        srt_str = ""

        for i, segment in enumerate(self.segments):
            # Convert start and end times to hh:mm:ss,sss format
            start_time = timedelta(seconds=segment.start)
            end_time = timedelta(seconds=segment.end)

            # Convert timedelta to appropriate string format
            str_start_time = str(start_time).zfill(8)
            if "." in str_start_time:
                str_start_time = str_start_time.replace(".", ",")
                if len(str_start_time.split(",")[1]) < 3:
                    str_start_time = str_start_time + "0" * (
                        3 - len(str_start_time.split(",")[1])
                    )
            else:
                str_start_time = str_start_time + ",000"

            str_end_time = str(end_time).zfill(8)
            if "." in str_end_time:
                str_end_time = str_end_time.replace(".", ",")
                if len(str_end_time.split(",")[1]) < 3:
                    str_end_time = str_end_time + "0" * (
                        3 - len(str_end_time.split(",")[1])
                    )
            else:
                str_end_time = str_end_time + ",000"

            # Add index, time range, and text to the string
            srt_str += f"{i+1}\n{str_start_time} --> {str_end_time}\n{segment.text.strip()}\n\n"

        if file_path:
            with open(file_path, "w") as f:
                f.write(srt_str)

        return srt_str

    # def segments(self) -> list[ShortSegment]:
    #     short_segments = []
    #     for segment in self.segments:
    #         short_segments.append(
    #             ShortSegment(
    #                 id=segment.id,
    #                 start=round(segment.start, 2),
    #                 end=round(segment.end, 2),
    #                 text=segment.text.strip(),
    #             )
    #         )
    #     return short_segments
